
/**
 * BoundedBuffer
 * Modification of BoundedBufferBase and ConditionalBoundedBuffer
 * Chapter 14 of "Java Concurrency in Practice"
 * 
 * @author Nicolás Wolovick
 */

public class BoundedQueue<T> {
    private final T[] items;
    private int tail;
    private int head;
    private int count;

    private boolean repInvariant() {
    	boolean r = 0<=count && count<=items.length;
    	// This could be extremely simple if a _sentinel_ buffer place were added,
    	// and a proper implementation of % were givem.
		if (tail>head)
			r = r && tail-head == count;
		else if (head>tail)
			r = r && tail-head+items.length == count;
		else
			r = r && (count==0 || count==items.length);
    	//if (!r)
    	//	System.out.println("c:"+count+" h:"+head+" t:"+tail);
    	return r;
	}

    @SuppressWarnings("unchecked")
	public BoundedQueue(int capacity) {
    	assert 0<capacity : "0<capacity";
    	tail = 0; head = 0; count = 0;
        this.items = (T[]) new Object[capacity];
    }

    public synchronized final void put(T x)  throws InterruptedException {
    	assert repInvariant() : "!RepInvariant";
        items[tail] = x;
        if (++tail == items.length)
            tail = 0;
        ++count;
        assert repInvariant() : "!RepInvariant";
    }

    public synchronized final T take() throws InterruptedException {
    	assert repInvariant() : "!RepInvariant";
        T v = items[head];
        items[head] = null;
        if (++head == items.length)
            head = 0;
        --count;
        assert repInvariant() : "!RepInvariant";
        return v;
    }

    public synchronized final boolean isFull() throws InterruptedException {
    	assert repInvariant() : "!RepInvariant";
        return count == items.length;
    }

    public synchronized final boolean isEmpty() throws InterruptedException {
    	assert repInvariant() : "!RepInvariant";
        return count == 0;
    }
}